#include "Resource.Shader.h"
#include "Toolkit.Display.h"
#include "Engine.File.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace res {

    const GLenum Shader::shType[] = { GL_FRAGMENT_SHADER, GL_VERTEX_SHADER };

    // Shader
    String Shader::unique( const String &Filename ) {
      return String().pushCChar( "Shader|filename=" ).pushString( Filename );
    }
    Shader* Shader::get( const String &Filename ) {
      Resource *R = global_s.Resources->find( unique( Filename ) );

      if ( !R )
        return new Shader( Filename );
      else
        return ( Shader* ) R;
    }

    Shader::Shader( const String &Filename ) : Resource( passError, unique( Filename ) ), Filename( String( "../res/shader/" ).pushString( Filename ) ) {
      Source = 0;
      PrgID = -1;

      for( SizeT a = 0; a < _tpLast; a ++ )
        ShdID[ a ] = -1;

      progress( pThreaded );
    }
    Shader::~Shader() {

    }

    void Shader::onLoad() {
      switch( Pass ) {

      case passLoad:
        if ( pLoad() )
          Pass = passError;
        break;

      case passCompile:
        if ( pCompile() )
          Pass = passError;
        break;

      case passError:
        break;

      default:
        ASSERT( true, "wrongPass" );
        break;

      }
    }
    void Shader::onUnload() {
      if( PrgID != -1 ) {
        glDeleteProgram( PrgID );
        PrgID = -1;
      }

      for( SizeT a = 0; a < _tpLast; a ++ ) {
        if ( ShdID[ a ] != -1 ) {
          glDeleteShader( ShdID[ a ] );
          ShdID[ a ] = -1;
        }
      }
    }

    bool Shader::pLoad() {
      if( glewIsSupported( "GL_VERSION_2_0" ) )
        ARB = false;
      else if( GLEW_ARB_vertex_shader && GLEW_ARB_fragment_shader )
        ARB = true;
      else
        ASSERT_THROW( true, "noShaderSupport" );

      Source.clear();
      readFile( Filename, Source );
      ASSERT_RETURNVAL( !Source.Size, true, "sourceReadFail", "filename", Filename.istr() );

      SizeT Pos = -1, SPos = -1;

      for( SizeT a = 0; a < Source.Size - 1; a ++ ) {
        if( Source[ a ] == '~' ) {
          SPos = a;
          break;
        }
      }

      Doc.loadString( IString( Source.Ptr, SPos ) );

      ASSERT_RETURNVAL( SPos == -1, true, "shaderConfigMissing", "filename", &Filename );

      for( SizeT a = SPos + 1; a < Source.Size - 1; a ++ ) {
        if( Source[ a ] == '~' ) {
          Pos = a;
          break;
        }
      }

      if( Pos == -1 ) {
        ShdS[ tpFragment ].assign( Source.Ptr + SPos + 1, Source.Size - SPos - 1 );
      } else {
        Source[ Pos ] = 0;
        ShdS[ tpVertex ].assign( Source.Ptr + SPos + 1, Pos - SPos - 1 );
        ShdS[ tpFragment ].assign( Source.Ptr + Pos + 1, Source.Size - Pos - 1);
      }

      progress( pMain );
      return false;
    }
    bool Shader::pCompile() {
      if( ARB )
        return pCompileARB();

      char *Log;
      GLint BLen = 0, BErr = 0, Sz;

      PrgID  = glCreateProgram();

      for ( UInt8 a = 0; a < _tpLast; a ++ ) {
        if ( Source[ a ] ) {

          if ( ShdS[ a ] == 0 )
            continue;

          ShdID[ a ] = glCreateShader( shType[ a ] );

          Sz = ShdS[ a ].Size;
          glShaderSource( ShdID[ a ], 1, ( const GLchar ** ) ShdS[ a ].Ptr, &Sz );
          glCompileShader( ShdID[ a ] );

          glGetShaderiv( ShdID[ a ], GL_COMPILE_STATUS, &BErr );
          if ( BErr != GL_TRUE ) {
            glGetShaderiv( ShdID[ a ], GL_INFO_LOG_LENGTH , &BLen );   
            Log = ( char* ) malloc( BLen );
            glGetInfoLogARB( ShdID[ a ], BLen, 0, Log );
            ASSERT_RETURNVAL( true, true, "shaderCompileErr", "filename", &Filename, "error", &IString( Log ) );

          }

          glAttachShader( PrgID, ShdID[ a ] );

        }
      }

      glLinkProgram( PrgID );
      glGetProgramiv( PrgID, GL_LINK_STATUS, &BErr );
      if ( BErr != GL_TRUE ) {
        glGetShaderiv( PrgID, GL_INFO_LOG_LENGTH , &BLen );   
        Log = ( char* ) malloc( BLen );
        glGetInfoLogARB( PrgID, BLen, 0, Log );
        ASSERT_RETURNVAL( true, true, "shaderLinkErr", "filename", &Filename, "error", &IString( Log ) );

      }

      GLuint P;
      char * Ch;

      cfx::Node &N1 = Doc >> "Attributes";
      foreach_a( N1.Nodes ) {
        Ch = N1.Nodes[ a ]->Name.toCChar();
        P = glGetAttribLocation( PrgID, Ch );
        free( Ch );
        Attributes.pushItem( P );
      }

      cfx::Node &N2 = Doc >> "Uniforms";
      foreach_a( N2.Nodes ) {
        Ch = N2.Nodes[ a ]->Name.toCChar();
        P = glGetUniformLocation( PrgID, Ch );
        free( Ch );
        Uniforms.pushItem( P );
      }

      progress( pEnd );
      return false;
    }
    bool Shader::pCompileARB() {
      char *Log;
      GLint BLen = 0, BErr = 0, Sz;

      PrgID  = glCreateProgramObjectARB();

      for ( UInt8 a = 0; a < _tpLast; a ++ ) {
        if ( Source[ a ] ) {

          if ( ShdS[ a ] == 0 )
            continue;

          ShdID[ a ] = glCreateShaderObjectARB( shType[ a ] );

          Sz = ShdS[ a ].Size;
          glShaderSourceARB( ShdID[ a ], 1, ( const GLchar ** ) ShdS[ a ].Ptr, &Sz );
          glCompileShaderARB( ShdID[ a ] );

          glGetShaderiv( ShdID[ a ], GL_COMPILE_STATUS, &BErr );
          if ( BErr != GL_TRUE ) {
            glGetShaderiv( ShdID[ a ], GL_INFO_LOG_LENGTH , &BLen );   
            Log = ( char* ) malloc( BLen );
            glGetInfoLogARB( ShdID[ a ], BLen, 0, Log );
            ASSERT_RETURNVAL( true, true, "shaderCompileErr", "filename", &Filename, "error", &IString( Log ) );

          }

          glAttachObjectARB( PrgID, ShdID[ a ] );

        }
      }

      glLinkProgramARB( PrgID );
      glGetProgramiv( PrgID, GL_LINK_STATUS, &BErr );
      if ( BErr != GL_TRUE ) {
        glGetShaderiv( PrgID, GL_INFO_LOG_LENGTH , &BLen );   
        Log = ( char* ) malloc( BLen );
        glGetInfoLogARB( PrgID, BLen, 0, Log );
        ASSERT_RETURNVAL( true, true, "shaderLinkErr", "filename", &Filename, "error", &IString( Log ) );

      }

      GLuint P;
      char * Ch;

      cfx::Node &N1 = Doc >> "Attributes";
      foreach_a( N1.Nodes ) {
        Ch = N1.Nodes[ a ]->Name.toCChar();
        P = glGetAttribLocationARB( PrgID, Ch );
        free( Ch );
        Attributes.pushItem( P );
      }

      cfx::Node &N2 = Doc >> "Uniforms";
      foreach_a( N2.Nodes ) {
        Ch = N2.Nodes[ a ]->Name.toCChar();
        P = glGetUniformLocationARB( PrgID, Ch );
        free( Ch );
        Uniforms.pushItem( P );
      }

      progress( pEnd );
      return false;
    }

    void Shader::bind() {
      if( Pass != passDone )
        return;

      foreach_a( Attributes ) {
        glEnableVertexAttribArray( Attributes[ a ] );
      }

      glProgram( PrgID );
    }
    void Shader::unbind() {
      foreach_a( Attributes ) {
        glDisableVertexAttribArray( Attributes[ a ] );
      }

      glProgram( 0 );
    }

  }

}

#endif