#include "Resource.Texture.h"
#include "Toolkit.Display.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace res {

    UInt32 TexUniqueID = 0;

    // Texture
    String Texture::unique( const String &Filename, bool LinearMap ) {
      if( LinearMap )
        return String().pushCChar( "Texture|mapping=linear|filename=" ).pushString( Filename );
      else
        return String().pushCChar( "Texture|mapping=nearest|filename=" ).pushString( Filename );
    }
    Texture* Texture::get( const String &Filename, bool LinearMap /* = false */ ) {
      Resource *R = global_s.Resources->find( unique( Filename, LinearMap ) );

      if ( !R )
        return new Texture( Filename, LinearMap );
      else
        return ( Texture* ) R;
    }

    void Texture::saveRaw( const String& Filename, Vec2Screen Pos, Vec2Screen Size ) {
      UInt32 Cap = Size.X * Size.Y * 4;
      void *Data = malloc( Cap );

      glReadPixels( Pos.X, Pos.Y, Size.X, Size.Y, GL_RGBA, GL_UNSIGNED_BYTE, Data );

      char *Nam = Filename.toCChar();
      FILE *f = fopen( Nam, "wb" );
      free( Nam );

      ASSERT_RETURN( !f, "fileOpenWriteFail", "filename", &Filename );

      fwrite( Data, 1, Cap, f );
      fclose( f );
    }

    Texture::Texture( const String &Filename, bool LinearMap ) : Resource( passError, unique( Filename, LinearMap ) ), Filename( String( "../res/image/" ).pushString( Filename ) ), LinMap( LinearMap ) {
      Tex_SDL = 0;
      Tex_OGL = -1;

      CreationType = ctLoadFromFile;

      progress( pThreaded );
    }
    Texture::Texture( Vec2Screen Size, bool LinearMap /* = false */ ) : Resource( passError, unique( String().pushInt( ( int ) TexUniqueID++ ), LinearMap ) ), Filename( String( "@" ) ), Size( Size ), LinMap( LinearMap ) {
      Tex_SDL = 0;
      Tex_OGL = -1;

      CreationType = ctJustCreate;

      progress( pMain );
    }

    Texture::~Texture() {

    }

    void Texture::onLoad() {
      if( CreationType == ctLoadFromFile ) {

        switch( Pass ) {

        case passLoad:
          if ( pLoad() )
            Pass = passError;
          break;

        case passRender:
          if ( pRender() )
            Pass = passError;
          break;

        case passError:
          break;

        default:
          ASSERT( true, "wrongPass" );
          break;

        }

      } else {

        switch( Pass ) {

        case passCreate:
          if ( pCreate() )
            Pass = passError;
          break;

        case passError:
          break;

        default:
          ASSERT( true, "wrongPass" );
          break;

        }

      }
    }
    void Texture::onUnload() {
      if( Tex_SDL ) {
        SDL_FreeSurface( Tex_SDL );
        Tex_SDL = 0;
      }

      if( Tex_OGL != -1 ) {
        glDeleteTextures( 1, &Tex_OGL );
        Tex_OGL = -1;
      }
    }

    bool Texture::pLoad() {
      char *Fn = Filename.toCChar();
      Tex_SDL = IMG_Load( Fn );
      free( Fn );

      Fn = ( char* ) SDL_GetError();
      ASSERT_RETURNVAL( !Tex_SDL, true, "textureLoadFail", "filename", &Filename, "error", &IString( IMG_GetError() ) );

      Size.assign( Tex_SDL->w, Tex_SDL->h );

      progress( pMain );
      return false;
    }
    bool Texture::pRender() {
      UInt8 Channels = Tex_SDL->format->BytesPerPixel;
      UInt32 Format;

      //ASSERT_RETURNVAL( Tex_SDL->format->Rmask != 0x000000ff, true, "surfaceWrongFormat" );

      if ( Channels == 4 ) 
        Format = GL_RGBA;
      else if ( Channels == 3 )
        Format = GL_RGB;
      else
        ASSERT_RETURNVAL( true, true, "surfaceNoTrueColor" );

      glGenTextures( 1, &Tex_OGL );
      ASSERT_RETURNVAL( !Tex_OGL, true, "textureCreateFail" );

      glTexture( Tex_OGL );

      if( LinMap ) {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
      } else {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
      }

      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

      glTexImage2D( GL_TEXTURE_2D, 0, Channels, Tex_SDL->w, Tex_SDL->h, 0, Format, GL_UNSIGNED_BYTE, Tex_SDL->pixels );

      SDL_FreeSurface( Tex_SDL );
      Tex_SDL = 0;

      progress( pEnd );
      return false;
    }
    bool Texture::pCreate() {
      Pass = passRender; // Skip a pass

      glGenTextures( 1, &Tex_OGL );
      ASSERT_RETURNVAL( !Tex_OGL, true, "textureCreateFail" );

      glTexture( Tex_OGL );

      if( LinMap ) {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
      } else {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
      }

      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

      glTexImage2D( GL_TEXTURE_2D, 0, 4, Size.X, Size.Y, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0 );

      Tex_SDL = 0;

      progress( pEnd );
      return false;
    }

    void Texture::draw( const Vec2Screen Pos ) {
      if( Pass != passDone )
        return;

      glTextureDrawMode();
      glTexture( Tex_OGL );

      glVertexBuffer.pushRect( Pos, Size );
      glTexCoordBuffer.pushRect( 0, 1 );
      glColorBuffer.pushColorRect( clWhite );
    }
    void Texture::draw( const Vec2Screen Pos, const Vec2Screen ISize ) {
      if( Pass != passDone )
        return;

      glTextureDrawMode();
      glTexture( Tex_OGL );

      glVertexBuffer.pushRect( Pos, Size );
      glTexCoordBuffer.pushRect( 0, 1 );
      glColorBuffer.pushColorRect( clWhite );
    }

    void Texture::drawTiled( Vec2Screen Pos, Vec2Screen ISize, Vec2Screen Tiles ) {
      if( Pass != passDone )
        return;

      glTextureDrawMode();
      glTexture( Tex_OGL );

      glVertexBuffer.pushRect( Pos, ISize );
      glTexCoordBuffer.pushRect( 0, Tiles.cast< float >() );
      glColorBuffer.pushColorRect( clWhite );
    }

    void Texture::drawPart( Vec2Screen Pos, Vec2Screen ISize, Vec2Tex TPos, Vec2Tex TSize ) {
      if( Pass != passDone )
        return;

      glTextureDrawMode();
      glTexture( Tex_OGL );

      glVertexBuffer.pushRect( Pos, ISize );
      glTexCoordBuffer.pushRect( TPos, TSize );
      glColorBuffer.pushColorRect( clWhite );
    }
    void Texture::drawPart( Vec2Screen Pos, Vec2Screen ISize, Vec2Screen TPos, Vec2Screen TSize ) {
      if( Pass != passDone )
        return;

      glTextureDrawMode();
      glTexture( Tex_OGL );

      glVertexBuffer.pushRect( Pos, ISize );
      glTexCoordBuffer.pushRect( TPos.cast< float >() / Size.cast< float >(), TSize.cast< float >() / Size.cast< float >() );
      glColorBuffer.pushColorRect( clWhite );
    }

    void Texture::drawPartRotated( Vec2Screen Pos, Vec2Screen ISize, Vec2Screen TPos, Vec2Screen TSize, Vec2Screen Center, Float64 Angle ) {
      if( Pass != passDone )
        return;

      glTextureDrawMode();
      glTexture( Tex_OGL );

      glVertexBuffer.pushRect( Pos, ISize );
      glVertexBuffer.rotateLastNPoints( 6, Center, ( Float32 ) Angle );

      glTexCoordBuffer.pushRect( TPos.cast< float >() / Size.cast< float >(), TSize.cast< float >() / Size.cast< float >() );
      glColorBuffer.pushColorRect( clWhite );
    }

  }

}

#endif